@@ -18,7 +18,7 @@ from queue import SimpleQueue
 from threading import Thread
 from typing import Any
 from subprocess import Popen, PIPE, STDOUT, TimeoutExpired
-from collections.abc import Callable
+from collections.abc import Callable, Sequence
 
 from xpra.platform.gui import (
     get_window_min_size, get_window_max_size,
@@ -29,8 +29,9 @@ from xpra.exit_codes import ExitCode, ExitValue
 from xpra.common import WINDOW_NOT_FOUND, WINDOW_DECODE_SKIPPED, WINDOW_DECODE_ERROR, noerr
 from xpra.platform.paths import get_icon_filename, get_resources_dir, get_python_execfile_command
 from xpra.scripts.config import FALSE_OPTIONS
+from xpra.client.gui.window_border import WindowBorder
 from xpra.util.thread import start_thread
-from xpra.util.str_fn import std, bytestostr, memoryview_to_bytes
+from xpra.util.str_fn import std, bytestostr, strtobytes, memoryview_to_bytes
 from xpra.os_util import OSX, POSIX, gi_import
 from xpra.util.system import is_Ubuntu, is_Wayland
 from xpra.util.objects import typedict, make_instance
@@ -81,6 +82,7 @@ ICON_OVERLAY: int = envint("XPRA_ICON_OVERLAY", 50)
 ICON_SHRINKAGE: int = envint("XPRA_ICON_SHRINKAGE", 75)
 SAVE_WINDOW_ICONS: bool = envbool("XPRA_SAVE_WINDOW_ICONS", False)
 SAVE_CURSORS: bool = envbool("XPRA_SAVE_CURSORS", False)
+POLL_POINTER = envint("XPRA_POLL_POINTER", 0)
 
 DRAW_LOG_FMT = "process_draw: %7i %8s for window %3i, sequence %8i, %4ix%-4i at %4i,%-4i" \
                " using %6s encoding with options=%s"
@@ -168,6 +170,67 @@ def parse_window_size(v, attribute="max-size"):
     return None
 
 
+def show_border_help() -> None:
+    if not first_time("border-help"):
+        return
+    log.info(" border format: color[,size][:off]")
+    log.info("  eg: red,10")
+    log.info("  eg: ,5")
+    log.info("  eg: auto,5")
+    log.info("  eg: blue")
+
+
+def parse_border(border_str="", display_name="", warn=False) -> WindowBorder:
+    # ie: "auto,5:off"
+    from xpra.gtk.widget import color_parse
+    parts = [x.strip() for x in border_str.replace(",", ":").split(":", 2)]
+    color_str = parts[0]
+    if color_str.lower() in ("none", "no", "off", "0"):
+        return WindowBorder(False)
+    if color_str.lower() == "help":
+        show_border_help()
+        return WindowBorder(False)
+    if color_str in ("auto", ""):
+        from hashlib import sha256
+        m = sha256()
+        if display_name:
+            m.update(strtobytes(display_name))
+        color_str = "#%s" % m.hexdigest()[:6]
+        log(f"border color derived from {display_name}: {color_str}")
+    try:
+        color = color_parse(color_str)
+        assert color is not None
+    except Exception as e:
+        if warn:
+            log.warn(f"Warning: invalid border color specified '{color_str!r}'")
+            if str(e):
+                log.warn(" %s", e)
+            show_border_help()
+        color = color_parse("red")
+    alpha = 0.6
+    size = 4
+    enabled = parts[-1] != "off"
+    if enabled and len(parts) >= 2:
+        size_str = parts[1]
+        try:
+            size = int(size_str)
+        except Exception as e:
+            if warn:
+                log.warn(f"Warning: invalid border size specified {size_str!r}")
+                log.warn(f" {e}")
+                show_border_help()
+        if size <= 0:
+            log(f"border size is {size}, disabling it")
+            enabled = False
+            size = 0
+        if size >= 45:
+            log.warn(f"Warning: border size is too large: {size}, clipping it")
+            size = 45
+    border = WindowBorder(enabled, color.red / 65536.0, color.green / 65536.0, color.blue / 65536.0, alpha, size)
+    log("parse_border(%s)=%s", border_str, border)
+    return border
+
+
 class WindowClient(StubClientMixin):
     """
     Utility superclass for clients that handle windows:
@@ -197,13 +260,14 @@ class WindowClient(StubClientMixin):
 
         self.server_window_frame_extents: bool = False
         self.server_is_desktop: bool = False
-        self.server_window_states: tuple[str, ...] = ()
-        self.server_window_signals: tuple[str, ...] = ()
+        self.server_window_states: Sequence[str] = ()
+        self.server_window_signals: Sequence[str] = ()
 
         self.server_input_devices = None
         self.server_precise_wheel: bool = False
 
         self.input_devices = "auto"
+        self.border = WindowBorder(False)
         self.border_str = "no"
 
         self.overlay_image = None
@@ -217,7 +281,6 @@ class WindowClient(StubClientMixin):
         self.server_bell: bool = False
         self.bell_enabled: bool = False
 
-        self.border = None
         self.window_close_action: str = "forward"
         self.modal_windows: bool = True
 
@@ -236,6 +299,8 @@ class WindowClient(StubClientMixin):
         self.pointer_grabbed = None
         self._suspended_at: float = 0
         self._button_state = {}
+        self.poll_pointer_timer = 0
+        self.poll_pointer_position = -1, -1
 
     def init(self, opts) -> None:
         if opts.system_tray:
@@ -269,7 +334,7 @@ class WindowClient(StubClientMixin):
 
         self.border_str = opts.border
         if opts.border:
-            self.parse_border()
+            self.border = parse_border(self.border_str)
 
         # mouse wheel:
         mw = (opts.mousewheel or "").lower().replace("-", "").split(",")
@@ -318,9 +383,13 @@ class WindowClient(StubClientMixin):
                         self.overlay_image = Image.open(icon_filename)
         traylog("overlay_image=%s", self.overlay_image)
 
-    def parse_border(self):
-        # not implemented here (see bindings client)
-        pass
+    def setup_connection(self, conn):
+        conn = super().setup_connection(conn)
+        display_name = getattr(self, "display_desc", {}).get("display_name", "")
+        if display_name:
+            # now that we have display_desc, parse the border again:
+            self.border = parse_border(self.border_str, display_name)
+        return conn
 
     def run(self) -> ExitValue:
         # we decode pixel data in this thread
@@ -340,6 +409,7 @@ class WindowClient(StubClientMixin):
         # (cleaner and needed when we run embedded in the client launcher)
         self.destroy_all_windows()
         self.cancel_lost_focus_timer()
+        self.cancel_poll_pointer_timer()
         if dq:
             dq.put(None)
         dt = self._draw_thread
@@ -355,7 +425,7 @@ class WindowClient(StubClientMixin):
             modal = w._metadata.boolget("modal", False)
             w.set_modal(modal)
 
-    def set_windows_cursor(self, client_windows, new_cursor):
+    def set_windows_cursor(self, client_windows, new_cursor) -> None:
         raise NotImplementedError()
 
     def window_bell(self, window, device, percent, pitch, duration, bell_class, bell_id, bell_name) -> None:
@@ -440,6 +510,11 @@ class WindowClient(StubClientMixin):
         # input devices:
         self.server_input_devices = c.strget("input-devices")
         self.server_precise_wheel = c.boolget("wheel.precise", False)
+        if POLL_POINTER:
+            if is_Wayland():
+                log.warn("Warning: pointer polling is unlikely to work under Wayland")
+                log.warn(" and may cause problems")
+            self.poll_pointer_timer = GLib.timeout_add(POLL_POINTER, self.poll_pointer)
         return True
 
     ######################################################################
@@ -468,7 +543,7 @@ class WindowClient(StubClientMixin):
 
     def send_wheel_delta(self, device_id: int, wid: int, button: int, distance, pointer=None, props=None) -> float:
         modifiers = self.get_current_modifiers()
-        buttons: tuple[int, ...] = ()
+        buttons: Sequence[int] = ()
         mouselog("send_wheel_deltas%s precise wheel=%s, modifiers=%s, pointer=%s",
                  (device_id, wid, button, distance, pointer, props), self.server_precise_wheel, modifiers, pointer)
         if self.server_precise_wheel:
@@ -479,7 +554,7 @@ class WindowClient(StubClientMixin):
                           button, idist,
                           pointer, modifiers, buttons] + list((props or {}).values())
                 mouselog("send_wheel_delta(..) %s", packet)
-                self.send_positional(packet)
+                self.send_positional(*packet)
             return 0
         # server cannot handle precise wheel,
         # so we have to use discrete events,
@@ -515,7 +590,8 @@ class WindowClient(StubClientMixin):
         mouselog("wheel_event%s new deltas=%s,%s",
                  (device_id, wid, deltax, deltay), self.wheel_deltax, self.wheel_deltay)
 
-    def send_button(self, device_id, wid, button, pressed, pointer, modifiers, buttons, props) -> None:
+    def send_button(self, device_id: int, wid: int, button: int, pressed: bool,
+                    pointer, modifiers, buttons, props) -> None:
         pressed_state = self._button_state.get(button, False)
         if SKIP_DUPLICATE_BUTTON_EVENTS and pressed_state == pressed:
             mouselog("button action: unchanged state, ignoring event")
@@ -551,17 +627,27 @@ class WindowClient(StubClientMixin):
             if props:
                 packet += list(props.values())
         mouselog("button packet: %s", packet)
-        self.send_positional(packet)
+        self.send_positional(*packet)
 
     def scale_pointer(self, pointer) -> tuple[int, int]:
         # subclass may scale this:
         # return int(pointer[0]/self.xscale), int(pointer[1]/self.yscale)
         return round(pointer[0]), round(pointer[1])
 
-    def send_input_devices(self, fmt, input_devices) -> None:
+    def send_input_devices(self, fmt: str, input_devices: dict[int, dict[str, Any]]) -> None:
         assert self.server_input_devices
         self.send("input-devices", fmt, input_devices)
 
+    def poll_pointer(self) -> bool:
+        pos = self.get_mouse_position()
+        if pos != self.poll_pointer_position:
+            self.poll_pointer_position = pos
+            device_id = -1
+            wid = 0
+            mouselog(f"poll_pointer() updated position: {pos}")
+            self.send_mouse_position(device_id, wid, pos)
+        return True
+
     ######################################################################
     # cursor:
     def _process_cursor(self, packet: PacketType) -> None:
@@ -653,7 +739,7 @@ class WindowClient(StubClientMixin):
                 modifiers = self.get_current_modifiers()
                 button_packet = ["button-action", wid, button, pressed, (x, y), modifiers]
                 traylog("button_packet=%s", button_packet)
-                self.send_positional(button_packet)
+                self.send_positional(*button_packet)
                 tray.reconfigure()
 
         def tray_mouseover(x, y):
@@ -682,13 +768,13 @@ class WindowClient(StubClientMixin):
             if tray_widget:
                 do_tray_geometry(*args)
             else:
-                self.idle_add(do_tray_geometry, *args)
+                GLib.idle_add(do_tray_geometry, *args)
 
         def tray_exit(*args):
             traylog("tray_exit(%s)", args)
 
         title = metadata.strget("title")
-        tray_widget = self.make_system_tray(app_id, None, title, None,
+        tray_widget = self.make_system_tray(app_id, None, title, "",
                                             tray_geometry, tray_click, tray_mouseover, tray_exit)
         traylog("setup_system_tray%s tray_widget=%s", (client, app_id, wid, w, h, title), tray_widget)
         assert tray_widget, "could not instantiate a system tray for tray id %s" % wid
@@ -877,7 +963,7 @@ class WindowClient(StubClientMixin):
                     return twid
         return 0
 
-    def patch_OR_popup_transient_for(self, metadata: typedict):
+    def patch_OR_popup_transient_for(self, metadata: typedict) -> None:
         pid = metadata.intget("pid", 0)
         twid = metadata.intget("transient-for", 0)
         if is_Wayland():
@@ -918,8 +1004,8 @@ class WindowClient(StubClientMixin):
                              border, self.max_window_size, self.default_cursor_data, self.pixel_depth,
                              self.headerbar)
                 break
-            except Exception:
-                log.warn("failed to instantiate %s", cwc, exc_info=True)
+            except (RuntimeError, ValueError):
+                log.warn(f"Warning: failed to instantiate {cwc!r}", exc_info=True)
         if window is None:
             log.warn("no more options.. this window will not be shown, sorry")
             return None
@@ -1145,7 +1231,7 @@ class WindowClient(StubClientMixin):
         # subclasses that wish to implement the feature may override this method
         return None
 
-    def get_client_window_classes(self, _w, _h, _metadata, _override_redirect) -> tuple[type, ...]:
+    def get_client_window_classes(self, _w, _h, _metadata, _override_redirect) -> Sequence[type]:
         return (self.ClientWindowClass,)
 
     def _process_new_window(self, packet: PacketType) -> None:
@@ -1238,6 +1324,7 @@ class WindowClient(StubClientMixin):
         if window:
             window.move_resize(ax, ay, aw, ah, -1)
 
+    # noinspection PyUnreachableCode
     def window_close_event(self, wid: int) -> None:
         log("window_close_event(%s) close window action=%s", wid, self.window_close_action)
         if self.window_close_action == "forward":
@@ -1335,17 +1422,14 @@ class WindowClient(StubClientMixin):
             try:
                 log("destroy_all_windows() destroying %s / %s", wid, window)
                 self.destroy_window(wid, window)
-            except Exception:
-                pass
+            except (RuntimeError, ValueError):
+                log(f"destroy_all_windows() failed to destroy {window}", exc_info=True)
         self._id_to_window = {}
         self._window_to_id = {}
         # signal watchers should have been killed in destroy_window(),
         # make sure we don't leave any behind:
         for signalwatcher in tuple(self._signalwatcher_to_wids.keys()):
-            try:
-                kill_signalwatcher(signalwatcher)
-            except Exception:
-                log("destroy_all_windows() error killing signal watcher %s", signalwatcher, exc_info=True)
+            kill_signalwatcher(signalwatcher)
 
     ######################################################################
     # bell
@@ -1406,6 +1490,12 @@ class WindowClient(StubClientMixin):
             self.lost_focus_timer = 0
             GLib.source_remove(lft)
 
+    def cancel_poll_pointer_timer(self) -> None:
+        ppt = self.poll_pointer_timer
+        if ppt:
+            self.poll_pointer_timer = 0
+            GLib.source_remove(ppt)
+
     ######################################################################
     # grabs:
     def window_grab(self, wid: int, _window) -> None:
@@ -1512,7 +1602,7 @@ class WindowClient(StubClientMixin):
         self._draw_queue.put(packet)
 
     def send_damage_sequence(self, wid: int, packet_sequence: int, width: int, height: int,
-                             decode_time, message="") -> None:
+                             decode_time: int, message="") -> None:
         packet = "damage-sequence", packet_sequence, wid, width, height, decode_time, message
         drawlog("sending ack: %s", packet)
         self.send_now(*packet)
@@ -1544,7 +1634,7 @@ class WindowClient(StubClientMixin):
         if not window:
             # window is gone
 
-            def draw_cleanup():
+            def draw_cleanup() -> None:
                 if coding == "mmap":
                     assert self.mmap_enabled
                     from xpra.net.mmap import int_from_buffer
@@ -1556,18 +1646,17 @@ class WindowClient(StubClientMixin):
                     # will get a chance to run first (preserving the order)
                 self.send_damage_sequence(wid, packet_sequence, width, height, WINDOW_NOT_FOUND, "window not found")
 
-            self.idle_add(draw_cleanup)
+            GLib.idle_add(draw_cleanup)
             return
         # rename old encoding aliases early:
-        options = {}
+        options = typedict()
         if len(packet) > 10:
-            options = packet[10]
-        options = typedict(options)
+            options.update(packet[10])
         dtype = DRAW_TYPES.get(type(data), type(data))
         drawlog(DRAW_LOG_FMT, len(data), dtype, wid, packet_sequence, width, height, x, y, coding, options)
         start = monotonic()
 
-        def record_decode_time(success, message=""):
+        def record_decode_time(success: bool | int, message="") -> None:
             if success > 0:
                 end = monotonic()
                 decode_time = round(end * 1000 * 1000 - start * 1000 * 1000)
@@ -1592,17 +1681,17 @@ class WindowClient(StubClientMixin):
                          coding, self._draw_counter, packet_sequence)
             if PAINT_FAULT_TELL:
                 msg = f"fault injection for {coding} draw packet {self._draw_counter}, sequence no={packet_sequence}"
-                self.idle_add(record_decode_time, False, msg)
+                GLib.idle_add(record_decode_time, False, msg)
             return
         # we could expose this to the csc step? (not sure how this could be used)
         # if self.xscale!=1 or self.yscale!=1:
         #    options["client-scaling"] = self.xscale, self.yscale
         try:
-            window.draw_region(x, y, width, height, coding, data, rowstride,
-                               packet_sequence, options, [record_decode_time])
+            window.draw_region(x, y, width, height, coding, data, rowstride, options, [record_decode_time])
         except Exception as e:
-            drawlog.error("Error drawing on window %i", wid, exc_info=True)
-            self.idle_add(record_decode_time, False, str(e))
+            drawlog.error("Error drawing on window %i", wid)
+            drawlog.error(f" using encoding {coding} with {options=}", exc_info=True)
+            GLib.idle_add(record_decode_time, False, str(e))
             raise
 
     ######################################################################
